#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from .attrembedlist import AttrEmbedList


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class FamilyAttrEmbedList(AttrEmbedList):
    def __init__(self, dbstate, uistate, track, data, config_key):
        AttrEmbedList.__init__(self, dbstate, uistate, track, data, config_key)

    def get_editor(self):
        from .. import EditAttribute

        return EditAttribute

    def get_user_values(self):
        return self.dbstate.db.get_family_attribute_types()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
# "Families with incomplete events"
# -------------------------------------------------------------------------
class FamilyWithIncompleteEvent(Rule):
    """Families with incomplete events"""

    name = _("Families with incomplete events")
    description = _(
        "Matches people with missing date or " "place in an event of the family"
    )
    category = _("Event filters")

    def apply_to_one(self, db: Database, person: Person) -> bool:
        for family_handle in person.family_list:
            family = db.get_family_from_handle(family_handle)
            if family:
                for event_ref in family.event_ref_list:
                    if event_ref:
                        event = db.get_event_from_handle(event_ref.ref)
                        if not event.place:
                            return True
                        if not event.date:
                            return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....lib.nameorigintype import NameOriginType

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Person
from ....db import Database


# -------------------------------------------------------------------------
#
# HasNameOf
#
# -------------------------------------------------------------------------
class HasNameOf(Rule):
    """Rule that checks for full or partial name matches"""

    labels = [
        _("Given name:"),
        _("Full Family name:"),
        _("Title:", "person"),
        _("Suffix:"),
        _("Call Name:"),
        _("Nick Name:"),
        _("Prefix:"),
        _("Single Surname:"),
        _("Connector"),
        _("Patronymic:"),
        _("Family Nick Name:"),
    ]
    name = _("People with the <name>")
    description = _("Matches people with a specified (partial) name")
    category = _("General filters")
    allow_regex = True

    def apply_to_one(self, db: Database, person: Person) -> bool:
        for name in [person.primary_name] + person.alternate_names:
            if self.match_name(name):
                return True
        return False

    def match_name(self, name):
        if self.list[0] and not self.match_substring(0, name.first_name):
            return False
        elif self.list[1] and not self.match_substring(1, name.get_surname()):
            return False
        elif self.list[2] and not self.match_substring(2, name.title):
            return False
        elif self.list[3] and not self.match_substring(3, name.suffix):
            return False
        elif self.list[4] and not self.match_substring(4, name.call):
            return False
        elif self.list[5] and not self.match_substring(5, name.nick):
            return False
        elif self.list[10] and not self.match_substring(10, name.famnick):
            return False
        else:
            for surn in name.surname_list:
                if self.match_surname(surn):
                    return True
        return False

    def match_surname(self, surn):
        if self.list[6] and not self.match_substring(6, surn.prefix):
            return False
        if self.list[7] and not self.match_substring(7, surn.surname):
            return False
        if self.list[8] and not self.match_substring(8, surn.connector):
            return False
        if int(surn.origintype.value) == NameOriginType.PATRONYMIC:
            if self.list[9] and not self.match_substring(9, surn.surname):
                return False
        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
#               2009       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import GLib

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Name, Surname
from gramps.gen.errors import WindowActiveError
from ...ddtargets import DdTargets
from .namemodel import NameModel
from .embeddedlist import TEXT_COL, MARKUP_COL, ICON_COL
from .groupembeddedlist import GroupEmbeddedList


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class NameEmbedList(GroupEmbeddedList):
    _HANDLE_COL = 2
    _DND_TYPE = DdTargets.NAME
    _WORKGROUP = NameModel.ALTINDEX

    _MSG = {
        "add": _("Create and add a new name"),
        "del": _("Remove the existing name"),
        "edit": _("Edit the selected name"),
        "up": _("Move the selected name upwards"),
        "down": _("Move the selected name downwards"),
    }

    # index = column in model. Value =
    #  (name, sortcol in model, width, markup/text, weigth_col
    _column_names = [
        (_("Name"), -1, 250, TEXT_COL, NameModel.COL_FONTWEIGHT[0], None),
        (_("Type"), NameModel.COL_TYPE[0], 100, TEXT_COL, -1, None),
        None,
        None,
        (_("Group As"), NameModel.COL_GROUPAS[0], 100, TEXT_COL, -1, None),
        (_("Source"), NameModel.COL_HASSOURCE[0], 30, ICON_COL, -1, "gramps-source"),
        (_("Notes Preview"), NameModel.COL_NOTEPREVIEW[0], 250, TEXT_COL, -1, None),
        (_("Private"), NameModel.COL_PRIVATE[0], 30, ICON_COL, -1, "gramps-lock"),
    ]

    def __init__(self, dbstate, uistate, track, data, person, config_key, callback):
        """callback is the function to call when preferred name changes
        on the namelist"""
        self.data = data
        self.person = person
        self.callback = callback

        GroupEmbeddedList.__init__(
            self,
            dbstate,
            uistate,
            track,
            _("_Names"),
            NameModel,
            config_key,
            move_buttons=True,
        )
        self.tree.expand_all()

    def _cleanup_on_exit(self):
        """Unset all things that can block garbage collection.
        Finalize rest
        """
        self.person = None
        self.callback = None
        self.data = None

    def get_data(self):
        return ([self.person.get_primary_name()], self.data)

    def groups(self):
        """
        Return the (group key, group name)s in the order as given by get_data()
        """
        return ((None, NameModel.DEFNAME), (None, NameModel.ALTNAME))

    def column_order(self):
        """
        The columns to show as a tuple of tuples containing
        tuples (show/noshow, model column)
        """
        return ((1, 0), (1, 5), (1, 7), (1, 1), (1, 4), (1, 6))

    def get_popup_menu_items(self):
        if self._tmpgroup == self._WORKGROUP:
            return [
                (True, _("_Add"), self.add_button_clicked),
                (False, _("_Edit"), self.edit_button_clicked),
                (True, _("_Remove"), self.del_button_clicked),
                (True, _("Set as Preferred name"), self.name_button_clicked),
            ]
        else:
            return [
                (True, _("_Add"), self.add_button_clicked),
                (False, _("_Edit"), self.edit_button_clicked),
            ]

    def name_button_clicked(self, obj):
        name = self.get_selected()
        if name and name[1]:
            self.set_default_name(name[1])

    def set_default_name(self, name):
        pname = self.person.get_primary_name()
        self.person.set_primary_name(name)
        remove = [altname for altname in self.data if altname.is_equal(name)]
        list(map(self.data.remove, remove))
        # only non empty name should move to alternative names
        if not name.is_equal(Name()):
            self.data.append(pname)
        self.rebuild()
        self.callback()

    def update_defname(self):
        """
        callback from person editor if change to the preferred name happens
        """
        self.model.update_defname(self.person.get_primary_name())
        self.tree.expand_all()

    def add_button_clicked(self, obj):
        name = Name()
        # the editor requires a surname
        name.add_surname(Surname())
        name.set_primary_surname(0)
        try:
            from .. import EditName

            EditName(self.dbstate, self.uistate, self.track, name, self.add_callback)
        except WindowActiveError:
            pass

    def add_callback(self, name):
        data = self.get_data()[self._WORKGROUP]
        data.append(name)
        self.rebuild()
        GLib.idle_add(self.tree.scroll_to_cell, (self._WORKGROUP, len(data) - 1))

    def edit_button_clicked(self, obj):
        name = self.get_selected()
        if name and name[1] is not None:
            try:
                from .. import EditName

                if name[0] == NameModel.ALTINDEX:
                    EditName(
                        self.dbstate,
                        self.uistate,
                        self.track,
                        name[1],
                        self.edit_callback,
                    )
                elif name[0] == NameModel.DEFINDEX:
                    EditName(
                        self.dbstate,
                        self.uistate,
                        self.track,
                        name[1],
                        self.editdef_callback,
                    )
            except WindowActiveError:
                pass

    def edit_callback(self, name):
        self.rebuild()

    def editdef_callback(self, name):
        """
        callback after default name has changed
        """
        self.rebuild()
        self.callback()

    def dropnotworkgroup(self, row, obj):
        """
        Drop of obj on row that is not WORKGROUP
        """
        if row[0] == NameModel.DEFINDEX:
            # drop on default name
            self.set_default_name(obj)

    def move_away_work(self, row_from, row_to, obj):
        """
        move from the workgroup to a not workgroup
        we allow to change the default name like this
        """
        if row_from[0] == self._WORKGROUP and row_to[0] == NameModel.DEFINDEX:
            self.set_default_name(obj)

    def post_rebuild(self, prebuildpath):
        """
        Allow post rebuild specific handling.
        @param prebuildpath: path selected before rebuild, None if none
        @type prebuildpath: tree path
        """
        self.tree.expand_all()
        if prebuildpath is not None:
            self.selection.select_path(prebuildpath)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2010       Raphael Ackermann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Set
from ...types import PrimaryObjectHandle, PrimaryObject
from ...db import Database


# -------------------------------------------------------------------------
#
# HasIdOf
#
# -------------------------------------------------------------------------
class HasGrampsId(Rule):
    """Rule that checks for an object with a specific Gramps ID."""

    labels = [_("ID:")]
    name = "Object with <Id>"
    description = "Matches objects with a specified Gramps ID"
    category = _("General filters")
    selected_handles: Set[PrimaryObjectHandle] = set([])

    def apply_to_one(self, db: Database, obj: PrimaryObject) -> bool:
        """
        apply the rule on the obj.
        return true if the rule passes, false otherwise.
        """
        return obj.handle in self.selected_handles

    def reset(self):
        self.selected_handles.clear()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modiy
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------
import pickle
import os
from xml.sax.saxutils import escape
from time import strftime as strftime

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GdkPixbuf
from gi.repository import Pango
import cairo

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.lib import NoteType
from gramps.gen.datehandler import get_date
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.errors import WindowActiveError
from gramps.gen.constfunc import mac
from .display import display_help
from .managedwindow import ManagedWindow
from .glade import Glade
from .ddtargets import DdTargets
from .makefilter import make_filter
from .utils import is_right_click, no_match_primary_mask
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.config import config
from gramps.gui.widgets.multitreeview import MultiTreeView

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
WIKI_HELP_PAGE = "%s_-_Navigation" % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Using_the_Clipboard", "manual")
clipdb = None  # current db to avoid different transient dbs during db change

# -------------------------------------------------------------------------
#
# icons used in the object listing
#
# -------------------------------------------------------------------------

theme = Gtk.IconTheme.get_default()
LINK_PIC = theme.load_icon("stock_link", 16, 0)
OBJ2ICON = {
    "media": "gramps-media",
    "note": "gramps-notes",
    "person": "gramps-person",
    "place": "gramps-place",
    "address": "gramps-address",
    "attribute": "gramps-attribute",
    "event": "gramps-event",
    "family": "gramps-family",
    "location": "geo-place-link",
    "media": "gramps-media",
    "name": "geo-show-person",
    "repository": "gramps-repository",
    "source": "gramps-source",
    "citation": "gramps-citation",
    "text": "gramps-font",
    "url": "gramps-geo",
}


def obj2icon(target):
    return OBJ2ICON[target] if target in OBJ2ICON else None


ICONS = {}
for name, icon in OBJ2ICON.items():
    ICONS[name] = theme.load_icon(icon, 16, 0)


# -------------------------------------------------------------------------
#
# Local functions
#
# -------------------------------------------------------------------------
def map2class(target):
    _d_ = {
        "person-link": ClipPersonLink,
        "family-link": ClipFamilyLink,
        "personref": ClipPersonRef,
        "childref": ClipChildRef,
        "source-link": ClipSourceLink,
        "citation-link": ClipCitation,
        "repo-link": ClipRepositoryLink,
        "pevent": ClipEvent,
        "eventref": ClipEventRef,
        "media": ClipMediaObj,
        "mediaref": ClipMediaRef,
        "place-link": ClipPlace,
        "placeref": ClipPlaceRef,
        "note-link": ClipNote,
        "TEXT": ClipText,
    }
    return _d_[target] if target in _d_ else None


def obj2class(target):
    _d_ = {
        "Person": ClipPersonLink,
        "Family": ClipFamilyLink,
        "Source": ClipSourceLink,
        "Citation": ClipCitation,
        "Repository": ClipRepositoryLink,
        "Event": ClipEvent,
        "Media": ClipMediaObj,
        "Place": ClipPlace,
        "Note": ClipNote,
    }
    return _d_[target] if target in _d_ else None


OBJ2TARGET = {
    "Person": Gdk.atom_intern("person-link", False),
    "Family": Gdk.atom_intern("family-link", False),
    "Source": Gdk.atom_intern("source-link", False),
    "Citation": Gdk.atom_intern("citation-link", False),
    "Repository": Gdk.atom_intern("repo-link", False),
    "Event": Gdk.atom_intern("pevent", False),
    "Media": Gdk.atom_intern("media", False),
    "Place": Gdk.atom_intern("place-link", False),
    "Note": Gdk.atom_intern("note-link", False),
}


def obj2target(target):
    return OBJ2TARGET[target] if target in OBJ2TARGET else None


def model_contains(model, data):
    """
    Returns True if data is a row in model.
    """
    # check type and value
    # data[0] is type of drop item, data[1] is Clip object
    for row in model:
        if data[0] == "TEXT":
            same = (row[0] == data[0]) and (row[1]._value == data[1]._value)
        else:
            # FIXME: too restrictive, birth and death won't both copy
            same = (
                (row[0] == data[0])
                and (row[1]._title == data[1]._title)
                and (row[1]._handle == data[1]._handle)
                and (row[3] == data[3])
                and (row[4] == data[4])
            )
        if same:
            return True
    return False


# -------------------------------------------------------------------------
#
# wrapper classes to provide object specific listing in the ListView
#
# -------------------------------------------------------------------------
class ClipWrapper:
    UNAVAILABLE_ICON = "dialog-error"

    def __init__(self, obj):
        self._obj = obj
        self._pickle = obj
        self._type = _("Unknown")
        self._objclass = None
        self._handle = None
        self._title = _("Unavailable")
        self._value = _("Unavailable")
        self._dbid = clipdb.get_dbid()
        self._dbname = clipdb.get_dbname()

    def get_type(self):
        return self._type

    def get_title(self):
        return self._title

    def get_value(self):
        return self._value

    def get_dbname(self):
        return self._dbname

    def pack(self):
        """
        Return a byte string that can be packed in a GtkSelectionData
        structure
        """
        if not self.is_valid():
            data = list(pickle.loads(self._pickle))
            data[2] = {
                "_obj": self._obj,
                "_pickle": self._pickle,
                "_type": self._type,
                "_handle": self._handle,
                "_objclass": self._objclass,
                "_title": self._title,
                "_value": self._value,
                "_dbid": self._dbid,
                "_dbname": self._dbname,
            }
            return pickle.dumps(data)
        if isinstance(self._obj, bytes):
            return self._obj
        # don't know if this happens in Gramps, theoretically possible
        asuni = str(self._obj)
        return asuni.encode("utf-8")

    def is_valid(self):
        if self._objclass and obj2class(self._objclass):
            # we have a primary object
            data = pickle.loads(self._obj)
            handle = data[2]
            return clipdb.method("has_%s_handle", self._objclass)(handle)
        return True


class ClipHandleWrapper(ClipWrapper):
    def __init__(self, obj):
        super(ClipHandleWrapper, self).__init__(obj)
        # unpack object
        drag_type, idval, data, val = pickle.loads(obj)
        if isinstance(data, dict):
            self.set_data(data)
        else:
            self._handle = data

    def set_data(self, data):
        for item in data:
            setattr(self, item, data[item])


class ClipObjWrapper(ClipWrapper):
    def __init__(self, obj):
        super(ClipObjWrapper, self).__init__(obj)
        # unpack object
        drag_type, idval, self._obj, val = pickle.loads(obj)
        self._pickle = obj

    def pack(self):
        if not self.is_valid():
            data = list(pickle.loads(self._pickle))
            data[2] = {
                "_obj": self._obj,
                "_pickle": self._pickle,
                "_type": self._type,
                "_handle": self._handle,
                "_objclass": self._objclass,
                "_title": self._title,
                "_value": self._value,
                "_dbid": self._dbid,
                "_dbname": self._dbname,
            }
            return pickle.dumps(data)
        return self._pickle

    def is_valid(self):
        if self._obj is None:
            return False

        for clname, handle in self._obj.get_referenced_handles_recursively():
            if obj2class(clname):  # a class we care about (not tag)
                if not clipdb.method("has_%s_handle", clname)(handle):
                    return False

        return True


class ClipAddress(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.ADDRESS]
    DRAG_TARGET = DdTargets.ADDRESS
    ICON = ICONS["address"]

    def __init__(self, obj):
        super(ClipAddress, self).__init__(obj)
        self._type = _("Address")
        self.refresh()

    def refresh(self):
        if self._obj:
            self._title = get_date(self._obj)
            self._value = "%s %s %s %s" % (
                self._obj.get_street(),
                self._obj.get_city(),
                self._obj.get_state(),
                self._obj.get_country(),
            )


class ClipLocation(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.LOCATION]
    DRAG_TARGET = DdTargets.LOCATION
    ICON = ICONS["location"]

    def __init__(self, obj):
        super(ClipLocation, self).__init__(obj)
        self._type = _("Location")
        self.refresh()

    def refresh(self):
        self._value = "%s %s %s" % (
            self._obj.get_city(),
            self._obj.get_state(),
            self._obj.get_country(),
        )


class ClipEvent(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.EVENT]
    DRAG_TARGET = DdTargets.EVENT
    ICON = ICONS["event"]

    def __init__(self, obj):
        super(ClipEvent, self).__init__(obj)
        self._type = _("Event")
        self._objclass = "Event"
        self.refresh()

    def refresh(self):
        if self._handle:
            value = clipdb.get_event_from_handle(self._handle)
            if value:
                self._title = str(value.get_type())
                self._value = value.get_description()


class ClipPlace(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.PLACE_LINK]
    DRAG_TARGET = DdTargets.PLACE_LINK
    ICON = ICONS["place"]

    def __init__(self, obj):
        super(ClipPlace, self).__init__(obj)
        self._type = _("Place")
        self._objclass = "Place"
        self.refresh()

    def refresh(self):
        if self._handle:
            value = clipdb.get_place_from_handle(self._handle)
            if value:
                self._title = value.gramps_id
                self._value = place_displayer.display(clipdb, value)


class ClipNote(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.NOTE_LINK]
    DRAG_TARGET = DdTargets.NOTE_LINK
    ICON = ICONS["note"]

    def __init__(self, obj):
        super(ClipNote, self).__init__(obj)
        self._type = _("Note")
        self._objclass = "Note"
        self.refresh()

    def refresh(self):
        value = clipdb.get_note_from_handle(self._handle)
        if value:
            self._title = value.get_gramps_id()
            note = value.get().replace("\n", " ")
            # String must be unicode for truncation to work for non
            # ascii characters
            note = str(note)
            if len(note) > 80:
                self._value = note[:80] + "..."
            else:
                self._value = note


class ClipFamilyEvent(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.FAMILY_EVENT]
    DRAG_TARGET = DdTargets.FAMILY_EVENT
    ICON = ICONS["family"]

    def __init__(self, obj):
        super(ClipFamilyEvent, self).__init__(obj)
        self._type = _("Family Event")
        self.refresh()

    def refresh(self):
        if self._obj:
            self._title = str(self._obj.get_type())
            self._value = self._obj.get_description()


class ClipUrl(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.URL]
    DRAG_TARGET = DdTargets.URL
    ICON = ICONS["url"]

    def __init__(self, obj):
        super(ClipUrl, self).__init__(obj)
        self._type = _("Url")
        self.refresh()

    def refresh(self):
        if self._obj:
            self._title = self._obj.get_path()
            self._value = self._obj.get_description()


class ClipAttribute(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.ATTRIBUTE]
    DRAG_TARGET = DdTargets.ATTRIBUTE
    ICON = ICONS["attribute"]

    def __init__(self, obj):
        super(ClipAttribute, self).__init__(obj)
        self._type = _("Attribute")
        self.refresh()

    def refresh(self):
        self._title = str(self._obj.get_type())
        self._value = self._obj.get_value()


class ClipFamilyAttribute(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.FAMILY_ATTRIBUTE]
    DRAG_TARGET = DdTargets.FAMILY_ATTRIBUTE
    ICON = ICONS["attribute"]

    def __init__(self, obj):
        super(ClipFamilyAttribute, self).__init__(obj)
        self._type = _("Family Attribute")
        self.refresh()

    def refresh(self):
        if self._obj:
            self._title = str(self._obj.get_type())
            self._value = self._obj.get_value()


class ClipCitation(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.CITATION_LINK]
    DRAG_TARGET = DdTargets.CITATION_LINK
    ICON = ICONS["citation"]

    def __init__(self, obj):
        super(ClipCitation, self).__init__(obj)
        self._type = _("Citation")
        self._objclass = "Citation"
        self.refresh()

    def refresh(self):
        if self._handle:
            try:
                citation = clipdb.get_citation_from_handle(self._handle)
                if citation:
                    self._title = citation.get_gramps_id()
                    notelist = list(
                        map(clipdb.get_note_from_handle, citation.get_note_list())
                    )
                    srctxtlist = [
                        note
                        for note in notelist
                        if note.get_type() == NoteType.SOURCE_TEXT
                    ]
                    page = citation.get_page()
                    if not page:
                        page = _("NA", "not available")
                    text = ""
                    if srctxtlist:
                        text = " ".join(srctxtlist[0].get().split())
                        # String must be unicode for truncation to work for non
                        # ascii characters
                        text = str(text)
                        if len(text) > 60:
                            text = text[:60] + "..."
                    self._value = _("Volume/Page: %(pag)s -- %(sourcetext)s") % {
                        "pag": page,
                        "sourcetext": text,
                    }
            except:
                # We are in the Source tree view. The shortcuts only
                # work for citations.
                print("We cannot copy the source from this view." " Use drag and drop.")
                self._title = self._value = ""
                self._pickle = self._type = self._objclass = None
                self._handle = self._dbid = self._dbname = None


class ClipRepoRef(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.REPOREF]
    DRAG_TARGET = DdTargets.REPOREF
    ICON = LINK_PIC

    def __init__(self, obj):
        super(ClipRepoRef, self).__init__(obj)
        self._type = _("Repository ref")
        self.refresh()

    def refresh(self):
        if self._obj:
            base = clipdb.get_repository_from_handle(self._obj.ref)
            if base:
                self._title = str(base.get_type())
                self._value = base.get_name()


class ClipEventRef(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.EVENTREF]
    DRAG_TARGET = DdTargets.EVENTREF
    ICON = LINK_PIC

    def __init__(self, obj):
        super(ClipEventRef, self).__init__(obj)
        self._type = _("Event ref")
        self.refresh()

    def refresh(self):
        if self._obj:
            base = clipdb.get_event_from_handle(self._obj.ref)
            if base:
                self._title = base.gramps_id
                self._value = str(base.get_type())


class ClipPlaceRef(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.PLACEREF]
    DRAG_TARGET = DdTargets.PLACEREF
    ICON = LINK_PIC

    def __init__(self, obj):
        super(ClipPlaceRef, self).__init__(obj)
        self._type = _("Place ref")
        self.refresh()

    def refresh(self):
        if self._obj:
            base = clipdb.get_place_from_handle(self._obj.ref)
            if base:
                self._title = base.gramps_id
                self._value = place_displayer.display(clipdb, base)


class ClipName(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.NAME]
    DRAG_TARGET = DdTargets.NAME
    ICON = ICONS["name"]

    def __init__(self, obj):
        super(ClipName, self).__init__(obj)
        self._type = _("Name")
        self.refresh()

    def refresh(self):
        if self._obj:
            self._title = str(self._obj.get_type())
            self._value = self._obj.get_name()


class ClipPlaceName(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.PLACENAME]
    DRAG_TARGET = DdTargets.PLACENAME
    ICON = ICONS["name"]

    def __init__(self, obj):
        super(ClipPlaceName, self).__init__(obj)
        self._type = _("Place Name")
        self.refresh()

    def refresh(self):
        if self._obj:
            self._title = get_date(self._obj)
            self._value = self._obj.get_value()


class ClipSurname(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.SURNAME]
    DRAG_TARGET = DdTargets.SURNAME
    ICON = ICONS["name"]

    def __init__(self, obj):
        super(ClipSurname, self).__init__(obj)
        self._type = _("Surname")
        self.refresh()

    def refresh(self):
        if self._obj:
            self._title = self._obj.get_surname()
            self._value = self._obj.get_surname()


class ClipText(ClipWrapper):
    DROP_TARGETS = DdTargets.all_text()
    DRAG_TARGET = DdTargets.TEXT
    ICON = ICONS["text"]

    def __init__(self, obj):
        super(ClipText, self).__init__(obj)
        self._type = _("Text")
        if isinstance(self._obj, bytes):
            self._pickle = str(self._obj, "utf-8")
        else:
            self._pickle = self._obj
        self.refresh()

    def refresh(self):
        self._title = _("Text")
        if isinstance(self._obj, bytes):
            self._value = str(self._obj, "utf-8")
        else:
            self._value = self._obj


class ClipMediaObj(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.MEDIAOBJ]
    DRAG_TARGET = DdTargets.MEDIAOBJ
    ICON = ICONS["media"]

    def __init__(self, obj):
        super(ClipMediaObj, self).__init__(obj)
        self._type = _("Media")
        self._objclass = "Media"
        self.refresh()

    def refresh(self):
        if self._handle:
            obj = clipdb.get_media_from_handle(self._handle)
            if obj:
                self._title = obj.get_description()
                self._value = obj.get_path()


class ClipMediaRef(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.MEDIAREF]
    DRAG_TARGET = DdTargets.MEDIAREF
    ICON = LINK_PIC

    def __init__(self, obj):
        super(ClipMediaRef, self).__init__(obj)
        self._type = _("Media ref")
        self.refresh()

    def refresh(self):
        if self._obj:
            base = clipdb.get_media_from_handle(self._obj.get_reference_handle())
            if base:
                self._title = base.get_description()
                self._value = base.get_path()


class ClipPersonRef(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.PERSONREF]
    DRAG_TARGET = DdTargets.PERSONREF
    ICON = LINK_PIC

    def __init__(self, obj):
        super(ClipPersonRef, self).__init__(obj)
        self._type = _("Person ref")
        self.refresh()

    def refresh(self):
        if self._obj:
            person = clipdb.get_person_from_handle(self._obj.get_reference_handle())
            if person:
                self._title = self._obj.get_relation()
                self._value = person.get_primary_name().get_name()


class ClipChildRef(ClipObjWrapper):
    DROP_TARGETS = [DdTargets.CHILDREF]
    DRAG_TARGET = DdTargets.CHILDREF
    ICON = LINK_PIC

    def __init__(self, obj):
        super(ClipChildRef, self).__init__(obj)
        self._type = _("Child ref")
        self.refresh()

    def refresh(self):
        if self._obj:
            person = clipdb.get_person_from_handle(self._obj.get_reference_handle())
            if person:
                frel = str(self._obj.get_father_relation())
                mrel = str(self._obj.get_mother_relation())
                self._title = _("%(frel)s %(mrel)s") % {"frel": frel, "mrel": mrel}
                self._value = person.get_primary_name().get_name()


class ClipPersonLink(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.PERSON_LINK]
    DRAG_TARGET = DdTargets.PERSON_LINK
    ICON = ICONS["person"]

    def __init__(self, obj):
        super(ClipPersonLink, self).__init__(obj)
        self._type = _("Person")
        self._objclass = "Person"
        self.refresh()

    def refresh(self):
        if self._handle:
            person = clipdb.get_person_from_handle(self._handle)
            if person:
                self._title = person.gramps_id
                self._value = person.get_primary_name().get_name()


class ClipFamilyLink(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.FAMILY_LINK]
    DRAG_TARGET = DdTargets.FAMILY_LINK
    ICON = ICONS["family"]

    def __init__(self, obj):
        super(ClipFamilyLink, self).__init__(obj)
        self._type = _("Family")
        self._objclass = "Family"
        self.refresh()

    def refresh(self):
        from gramps.gen.simple import SimpleAccess

        if self._handle:
            family = clipdb.get_family_from_handle(self._handle)
            if family:
                _sa = SimpleAccess(clipdb)
                self._title = family.gramps_id
                self._value = _sa.describe(family)


class ClipSourceLink(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.SOURCE_LINK]
    DRAG_TARGET = DdTargets.SOURCE_LINK
    ICON = ICONS["source"]

    def __init__(self, obj):
        super(ClipSourceLink, self).__init__(obj)
        self._type = _("Source")
        self._objclass = "Source"
        self.refresh()

    def refresh(self):
        if self._handle:
            source = clipdb.get_source_from_handle(self._handle)
            if source:
                self._title = source.get_gramps_id()
                self._value = source.get_title()


class ClipRepositoryLink(ClipHandleWrapper):
    DROP_TARGETS = [DdTargets.REPO_LINK]
    DRAG_TARGET = DdTargets.REPO_LINK
    ICON = ICONS["repository"]

    def __init__(self, obj):
        super(ClipRepositoryLink, self).__init__(obj)
        self._type = _("Repository")
        self._objclass = "Repository"
        self.refresh()

    def refresh(self):
        if self._handle:
            source = clipdb.get_repository_from_handle(self._handle)
            if source:
                self._title = str(source.get_type())
                self._value = source.get_name()


# -------------------------------------------------------------------------
#
# Wrapper classes to deal with lists of objects
#
# -------------------------------------------------------------------------


class ClipDropList:
    DROP_TARGETS = [DdTargets.LINK_LIST]
    DRAG_TARGET = None

    def __init__(self, obj_list):
        # ('link-list', id, (('person-link', handle),
        #                    ('person-link', handle), ...), 0)
        self._obj_list = pickle.loads(obj_list)

    def get_objects(self):
        list_type, _id, handles, timestamp = self._obj_list
        retval = []
        for target, handle in handles:
            _class = map2class(target)
            if _class:
                obj = _class(pickle.dumps((target, _id, handle, timestamp)))
                if obj:
                    retval.append(obj)
        return retval


class ClipDropRawList(ClipDropList):
    DROP_TARGETS = [DdTargets.RAW_LIST]
    DRAG_TARGET = None

    def __init__(self, obj_list):
        # ('raw-list', id, (ClipObject, ClipObject, ...), 0)
        self._obj_list = pickle.loads(obj_list)

    def get_objects(self):
        retval = []
        for item in self._obj_list:
            if item is None:
                continue
            target = pickle.loads(item)[0]
            _class = map2class(target)
            if _class:
                obj = _class(item)
                if obj:
                    retval.append(obj)
        return retval


class ClipDropHandleList(ClipDropList):
    DROP_TARGETS = [DdTargets.HANDLE_LIST]
    DRAG_TARGET = None

    def __init__(self, obj_list):
        # incoming:
        # ('handle-list', id, (('Person', '2763526751235'),
        #                      ('Source', '3786234743978'), ...), 0)
        self._obj_list = pickle.loads(obj_list)

    def get_objects(self):
        retval = []
        for objclass, handle in self._obj_list:
            _class = obj2class(objclass)
            target = obj2target(objclass).name()
            # outgoing:
            # (drag_type, idval, self._handle, val) = pickle.loads(self._obj)
            data = (target, id(self), handle, 0)
            obj = _class(pickle.dumps(data))
            retval.append(obj)
        return retval


# FIXME: add family


# -------------------------------------------------------------------------
#
# ClipboardListModel class
#
# -------------------------------------------------------------------------
class ClipboardListModel(Gtk.ListStore):
    def __init__(self):
        Gtk.ListStore.__init__(
            self,
            str,  # 0: object type
            object,  # 1: object
            object,  # 2: tooltip callback
            str,  # 3: type
            str,  # 4: value
            str,  # 5: unique database id (dbid)
            str,
        )  # 6: db name (may be old)


# -------------------------------------------------------------------------
#
# ClipboardListView class
#
# -------------------------------------------------------------------------
class ClipboardListView:
    LOCAL_DRAG_TYPE = "MY_TREE_MODEL_ROW"
    LOCAL_DRAG_TARGET = Gtk.TargetEntry.new(
        LOCAL_DRAG_TYPE, Gtk.TargetFlags.SAME_WIDGET, 0
    )

    def __init__(self, dbstate, widget):
        self._widget = widget
        self.dbstate = dbstate
        self.dbstate.connect("database-changed", self.database_changed)
        self.database_changed(dbstate.db)

        self._target_type_to_wrapper_class_map = {}
        self._previous_drop_time = 0

        # Create the tree columns
        self._col1 = Gtk.TreeViewColumn(_("Type"))
        self._col1.set_property("resizable", True)
        self._col1.set_sort_column_id(0)
        self._col2 = Gtk.TreeViewColumn(_("Title"))
        self._col2.set_property("resizable", True)
        self._col2.set_sort_column_id(3)
        self._col3 = Gtk.TreeViewColumn(_("Value"))
        self._col3.set_property("resizable", True)
        self._col3.set_sort_column_id(4)
        self._col4 = Gtk.TreeViewColumn(_("Family Tree"))
        self._col4.set_property("resizable", True)
        self._col4.set_sort_column_id(6)

        # Add columns
        self._widget.append_column(self._col1)
        self._widget.append_column(self._col2)
        self._widget.append_column(self._col3)
        self._widget.append_column(self._col4)

        # Create cell renders
        self._col1_cellpb = Gtk.CellRendererPixbuf()
        self._col1_cell = Gtk.CellRendererText()
        self._col2_cell = Gtk.CellRendererText()
        self._col3_cell = Gtk.CellRendererText()
        self._col4_cell = Gtk.CellRendererText()

        # Add cells to view
        self._col1.pack_start(self._col1_cellpb, False)
        self._col1.pack_start(self._col1_cell, True)
        self._col2.pack_start(self._col2_cell, True)
        self._col3.pack_start(self._col3_cell, True)
        self._col4.pack_start(self._col4_cell, True)

        # Setup the cell data callback funcs
        self._col1.set_cell_data_func(self._col1_cellpb, self.object_pixbuf)
        self._col1.set_cell_data_func(self._col1_cell, self.object_type)
        self._col2.set_cell_data_func(self._col2_cell, self.object_title)
        self._col3.set_cell_data_func(self._col3_cell, self.object_value)
        self._col4.set_cell_data_func(self._col4_cell, self.get_dbname)

        # Set the column that inline searching will use.
        self._widget.set_enable_search(True)
        # self._widget.set_search_column(3)

        targ_data = (ClipboardListView.LOCAL_DRAG_TARGET,) + DdTargets.all_targets()
        self._widget.drag_dest_set(Gtk.DestDefaults.ALL, targ_data, Gdk.DragAction.COPY)

        self._widget.connect("drag-data-get", self.object_drag_data_get)
        self._widget.connect("drag-begin", self.object_drag_begin)
        self._widget.connect_after("drag-begin", self.object_after_drag_begin)
        self._widget.connect("drag-data-received", self.object_drag_data_received)
        self._widget.connect("drag-end", self.object_drag_end)

        self.register_wrapper_classes()

    def database_changed(self, db):
        if not db.is_open():
            return
        global clipdb
        clipdb = db
        # Note: delete event is emitted before the delete, so checking
        #        if valid on this is useless !
        db_signals = (
            "person-update",
            "person-rebuild",
            "family-update",
            "family-rebuild",
            "source-update",
            "source-rebuild",
            "place-update",
            "place-rebuild",
            "media-update",
            "media-rebuild",
            "event-update",
            "event-rebuild",
            "repository-update",
            "repository-rebuild",
            "note-rebuild",
        )

        for signal in db_signals:
            clipdb.connect(signal, self.refresh_objects)

        clipdb.connect("person-delete", gen_del_obj(self.delete_object, "person-link"))
        clipdb.connect(
            "person-delete", gen_del_obj(self.delete_object_ref, "personref")
        )
        clipdb.connect("person-delete", gen_del_obj(self.delete_object_ref, "childref"))
        clipdb.connect("source-delete", gen_del_obj(self.delete_object, "source-link"))
        clipdb.connect("source-delete", gen_del_obj(self.delete_object_ref, "srcref"))
        clipdb.connect(
            "repository-delete", gen_del_obj(self.delete_object, "repo-link")
        )
        clipdb.connect("event-delete", gen_del_obj(self.delete_object, "pevent"))
        clipdb.connect("event-delete", gen_del_obj(self.delete_object_ref, "eventref"))
        clipdb.connect("media-delete", gen_del_obj(self.delete_object, "media"))
        clipdb.connect("media-delete", gen_del_obj(self.delete_object_ref, "mediaref"))
        clipdb.connect("place-delete", gen_del_obj(self.delete_object, "place-link"))
        clipdb.connect("note-delete", gen_del_obj(self.delete_object, "note-link"))
        # family-delete not needed, cannot be dragged!

        self.refresh_objects()

    def refresh_objects(self, dummy=None):
        model = self._widget.get_model()

        if model:
            for _ob in model:
                if not _ob[1].is_valid():
                    model.remove(_ob.iter)
                else:
                    _ob[1].refresh()
                    _ob[4] = _ob[1].get_value()  # Force listview to update

    def delete_object(self, handle_list, link_type):
        model = self._widget.get_model()

        if model:
            for _ob in model:
                if _ob[0] == link_type:
                    data = pickle.loads(_ob[1]._obj)
                    if data[2] in handle_list:
                        model.remove(_ob.iter)

    def delete_object_ref(self, handle_list, link_type):
        model = self._widget.get_model()

        if model:
            for _ob in model:
                if _ob[0] == link_type:
                    data = _ob[1]._obj.get_reference_handle()
                    if data in handle_list:
                        model.remove(_ob.iter)

    # Method to manage the wrapper classes.

    def register_wrapper_classes(self):
        self.register_wrapper_class(ClipAddress)
        self.register_wrapper_class(ClipLocation)
        self.register_wrapper_class(ClipEvent)
        self.register_wrapper_class(ClipPlace)
        self.register_wrapper_class(ClipEventRef)
        self.register_wrapper_class(ClipPlaceRef)
        self.register_wrapper_class(ClipRepoRef)
        self.register_wrapper_class(ClipFamilyEvent)
        self.register_wrapper_class(ClipUrl)
        self.register_wrapper_class(ClipAttribute)
        self.register_wrapper_class(ClipFamilyAttribute)
        self.register_wrapper_class(ClipName)
        self.register_wrapper_class(ClipPlaceName)
        self.register_wrapper_class(ClipRepositoryLink)
        self.register_wrapper_class(ClipMediaObj)
        self.register_wrapper_class(ClipMediaRef)
        self.register_wrapper_class(ClipSourceLink)
        self.register_wrapper_class(ClipCitation)
        self.register_wrapper_class(ClipPersonLink)
        self.register_wrapper_class(ClipFamilyLink)
        self.register_wrapper_class(ClipDropList)
        self.register_wrapper_class(ClipDropRawList)
        self.register_wrapper_class(ClipDropHandleList)
        self.register_wrapper_class(ClipPersonRef)
        self.register_wrapper_class(ClipChildRef)
        self.register_wrapper_class(ClipText)
        self.register_wrapper_class(ClipNote)

    def register_wrapper_class(self, wrapper_class):
        for drop_target in wrapper_class.DROP_TARGETS:
            self._target_type_to_wrapper_class_map[drop_target.drag_type] = (
                wrapper_class
            )

    # Methods for rendering the cells.

    def object_pixbuf(self, column, cell, model, node, user_data=None):
        _ob = model.get_value(node, 1)
        if _ob._dbid is not None and _ob._dbid != self.dbstate.db.get_dbid():
            if isinstance(_ob.__class__.UNAVAILABLE_ICON, str):
                cell.set_property("icon-name", _ob.__class__.UNAVAILABLE_ICON)
            else:
                cell.set_property("pixbuf", _ob.__class__.UNAVAILABLE_ICON)
        else:
            cell.set_property("pixbuf", _ob.__class__.ICON)

    def object_type(self, column, cell, model, node, user_data=None):
        _ob = model.get_value(node, 1)
        cell.set_property("text", _ob.get_type())

    def object_title(self, column, cell, model, node, user_data=None):
        _ob = model.get_value(node, 1)
        cell.set_property("text", _ob.get_title())

    def object_value(self, column, cell, model, node, user_data=None):
        _ob = model.get_value(node, 1)
        cell.set_property("text", _ob.get_value())

    def get_dbname(self, column, cell, model, node, user_data=None):
        _ob = model.get_value(node, 1)
        cell.set_property("text", _ob.get_dbname())

    # handlers for the drag and drop events.

    def on_object_select_row(self, obj):
        tree_selection = self._widget.get_selection()
        model, paths = tree_selection.get_selected_rows()
        if len(paths) > 1:
            targets = [DdTargets.RAW_LIST.target(), ClipboardListView.LOCAL_DRAG_TARGET]
        else:
            targets = [ClipboardListView.LOCAL_DRAG_TARGET]
        for path in paths:
            node = model.get_iter(path)
            if node is not None:
                _ob = model.get_value(node, 1)
                targets += [target.target() for target in _ob.__class__.DROP_TARGETS]

        self._widget.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            targets,
            Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
        )

    def object_drag_begin(self, widget, drag_context):
        """Handle the beginning of a drag operation."""
        pass

    def object_after_drag_begin(self, widget, drag_context):
        tree_selection = widget.get_selection()
        model, paths = tree_selection.get_selected_rows()
        if len(paths) == 1:
            path = paths[0]
            node = model.get_iter(path)
            target = model.get_value(node, 0)
            if target == "TEXT":
                layout = widget.create_pango_layout(model.get_value(node, 4))
                layout.set_alignment(Pango.Alignment.CENTER)
                width, height = layout.get_pixel_size()
                surface = cairo.ImageSurface(cairo.FORMAT_RGB24, width, height)
                ctx = cairo.Context(surface)
                style_ctx = self._widget.get_style_context()
                Gtk.render_background(style_ctx, ctx, 0, 0, width, height)
                ctx.save()
                Gtk.render_layout(style_ctx, ctx, 0, 0, layout)
                ctx.restore()
                Gtk.drag_set_icon_surface(drag_context, surface)
            else:
                try:
                    if map2class(target):
                        Gtk.drag_set_icon_pixbuf(
                            drag_context, map2class(target).ICON, 0, 0
                        )
                except:
                    Gtk.drag_set_icon_default(drag_context)

    def object_drag_end(self, widget, drag_context):
        """Handle the end of a drag operation."""
        pass

    def object_drag_data_get(self, widget, context, sel_data, info, time):
        tree_selection = widget.get_selection()
        model, paths = tree_selection.get_selected_rows()
        tgs = context.list_targets()
        if len(paths) == 1:
            path = paths[0]
            node = model.get_iter(path)
            _ob = model.get_value(node, 1)
            if model.get_value(node, 0) == "TEXT":
                sel_data.set_text(_ob._value, -1)
            else:
                sel_data.set(tgs[0], 8, _ob.pack())
        elif len(paths) > 1:
            raw_list = []
            for path in paths:
                node = model.get_iter(path)
                _ob = model.get_value(node, 1)
                raw_list.append(_ob.pack())
            sel_data.set(tgs[0], 8, pickle.dumps(raw_list))

    def object_drag_data_received(
        self,
        widget,
        context,
        x,
        y,
        selection,
        info,
        time,
        title=None,
        value=None,
        dbid=None,
        dbname=None,
    ):
        model = widget.get_model()
        sel_data = selection.get_data()
        # In Windows time is always zero. Until that is fixed, use the seconds
        # of the local time to filter out double drops.
        real_time = strftime("%S")

        # There is a strange bug that means that if there is a selection
        # in the list we get multiple drops of the same object. Luckily
        # the time values are the same so we can drop all but the first.
        if (real_time == self._previous_drop_time) and (time != -1):
            return None

        # Find a wrapper class
        possible_wrappers = []
        if mac():
            # context is empty on mac due to a bug, work around this
            # Note that this workaround code works fine in linux too as
            # we know very well inside of Gramps what sel_data can be, so
            # we can anticipate on it, instead of letting the wrapper handle
            # it. This is less clean however !
            # See http://www.gramps-project.org/bugs/view.php?id=3089 for
            # an explaination of why this is required.
            dragtype = None
            try:
                dragtype = pickle.loads(sel_data)[0]
            except pickle.UnpicklingError as msg:
                # not a pickled object, probably text
                if isinstance(sel_data, str):
                    dragtype = DdTargets.TEXT.drag_type
            if dragtype in self._target_type_to_wrapper_class_map:
                possible_wrappers = [dragtype]
        else:
            tgs = [atm.name() for atm in context.list_targets()]
            possible_wrappers = [
                target
                for target in tgs
                if target in self._target_type_to_wrapper_class_map
            ]

        if not possible_wrappers:
            # No wrapper for this class
            return None

        # Just select the first match.
        wrapper_class = self._target_type_to_wrapper_class_map[
            str(possible_wrappers[0])
        ]
        try:
            _ob = wrapper_class(sel_data)
            if title:
                _ob._title = title
            if value:
                _ob._value = value
            if dbid:
                _ob._dbid = dbid
            if dbname:
                _ob._dbname = dbname

            # If this was a ClipPersonLink with a list for a handle, we don't
            # want it.  Can happen in People Treeview with attempt to drag last
            # name group.
            if isinstance(_ob, ClipHandleWrapper) and isinstance(_ob._handle, list):
                return None
            # If the wrapper object is a subclass of ClipDropList then
            # the drag data was a list of objects and we need to decode
            # all of them.
            if isinstance(_ob, ClipDropList):
                o_list = _ob.get_objects()
            else:
                o_list = [_ob]
            for _ob in o_list:
                if _ob.__class__.DRAG_TARGET is None:
                    continue
                data = [
                    _ob.__class__.DRAG_TARGET.drag_type,
                    _ob,
                    None,
                    _ob._type,
                    _ob._value,
                    _ob._dbid,
                    _ob._dbname,
                ]
                contains = model_contains(model, data)
                if contains and not (context.get_actions() & Gdk.DragAction.MOVE):
                    continue
                drop_info = widget.get_dest_row_at_pos(x, y)
                if drop_info:
                    path, position = drop_info
                    node = model.get_iter(path)
                    if (
                        position == Gtk.TreeViewDropPosition.BEFORE
                        or position == Gtk.TreeViewDropPosition.INTO_OR_BEFORE
                    ):
                        model.insert_before(node, data)
                    else:
                        model.insert_after(node, data)
                elif isinstance(data[1], ClipCitation):
                    if data[3]:
                        # we have a real citation
                        model.append(data)
                    # else:
                    #   We are in a Source treeview and trying
                    #   to copy a source with a shortcut.
                    #   Use drag and drop to do that.
                else:
                    model.append(data)

            # FIXME: there is one bug here: if you multi-select and drop
            # on self, then it moves the first, and copies the rest.

            if context.get_actions() & Gdk.DragAction.MOVE:
                context.finish(True, True, time)

            # remember time for double drop workaround.
            self._previous_drop_time = real_time
            return o_list
        except EOFError:
            return None

    # proxy methods to provide access to the real widget functions.

    def set_model(self, model=None):
        self._widget.set_model(model)
        self._widget.get_selection().connect("changed", self.on_object_select_row)
        self._widget.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

    def get_model(self):
        return self._widget.get_model()

    def get_selection(self):
        return self._widget.get_selection()

    def set_search_column(self, col):
        return self._widget.set_search_column(col)


# -------------------------------------------------------------------------
#
# ClipboardWindow class
#
# -------------------------------------------------------------------------
class ClipboardWindow(ManagedWindow):
    """
    The Clipboard provides a temporary area to hold objects that can
    be reused accross multiple Person records. The pad provides a window
    onto which objects can be dropped and then dragged into new Person
    dialogs. The objects are stored as the pickles that are built by the
    origininating widget. The objects are only unpickled in order to
    provide the text in the display.

    No attempt is made to ensure that any references contained within
    the pickles are valid. Because the pad extends the life time of drag
    and drop objects, it is possible that references that were valid
    when an object is copied to the pad are invalid by the time they
    are dragged to a new Person. For this reason, using the pad places
    a responsibility on all '_drag_data_received' methods to check the
    references of objects before attempting to use them.
    """

    # Class attribute used to hold the content of the Clipboard.
    # A class attribute is used so that the content
    # it preserved even when the Clipboard window is closed.
    # As there is only ever one Clipboard we do not need to
    # maintain a list of these.
    otree = None

    def __init__(self, dbstate, uistate):
        """Initialize the ClipboardWindow class, and display the window"""

        ManagedWindow.__init__(self, uistate, [], self.__class__)
        self.dbstate = dbstate

        self.database_changed(self.dbstate.db)
        self.dbstate.connect("database-changed", self.database_changed)

        self.top = Glade()
        self.set_window(self.top.toplevel, None, None, msg=_("Clipboard"))
        self.setup_configs("interface.clipboard", 500, 300)
        if not config.get("behavior.immediate-warn"):
            self.get_window().set_tooltip_text(_("Any changes are saved immediately"))

        self.clear_all_btn = self.top.get_object("btn_clear_all")
        self.clear_btn = self.top.get_object("btn_clear")
        objectlist = self.top.get_object("objectlist")
        mtv = ClipboardMultiTreeView(self.dbstate, self.uistate, _("Clipboard"))
        scrolledwindow = self.top.get_object("scrolledwindow86")
        scrolledwindow.remove(objectlist)
        scrolledwindow.add(mtv)
        self.object_list = ClipboardListView(self.dbstate, mtv)
        self.object_list.get_selection().connect(
            "changed", self.set_clear_btn_sensitivity
        )
        self.object_list.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        self.set_clear_btn_sensitivity(sel=self.object_list.get_selection())

        if not ClipboardWindow.otree:
            ClipboardWindow.otree = ClipboardListModel()

        self.set_clear_all_btn_sensitivity(treemodel=ClipboardWindow.otree)
        ClipboardWindow.otree.connect("row-deleted", self.set_clear_all_btn_sensitivity)
        ClipboardWindow.otree.connect(
            "row-inserted", self.set_clear_all_btn_sensitivity
        )

        self.object_list.set_model(ClipboardWindow.otree)

        # Database might have changed, objects might have been removed,
        # we need to reevaluate if all data is valid
        self.object_list.refresh_objects()

        self.top.connect_signals(
            {
                "on_close_clipboard": self.close,
                "on_clear_clicked": self.on_clear_clicked,
                "on_help_clicked": self.on_help_clicked,
            }
        )

        self.clear_all_btn.connect_object(
            "clicked", Gtk.ListStore.clear, ClipboardWindow.otree
        )
        self.db.connect("database-changed", lambda x: ClipboardWindow.otree.clear())

        mtv.restore_column_size()
        self.show()

    def build_menu_names(self, obj):
        return (_("Clipboard"), None)

    def database_changed(self, database):
        self.db = database

    def set_clear_all_btn_sensitivity(
        self, treemodel=None, path=None, node=None, user_param1=None
    ):
        if treemodel:
            self.clear_all_btn.set_sensitive(True)
        else:
            self.clear_all_btn.set_sensitive(False)

    def set_clear_btn_sensitivity(self, sel=None, user_param1=None):
        if sel.count_selected_rows() == 0:
            self.clear_btn.set_sensitive(False)
        else:
            self.clear_btn.set_sensitive(True)

    def on_help_clicked(self, obj):
        """Display the relevant portion of Gramps manual"""
        display_help(webpage=WIKI_HELP_PAGE, section=WIKI_HELP_SEC)

    def on_clear_clicked(self, obj):
        """Deletes the selected object from the object list"""
        selection = self.object_list.get_selection()
        model, paths = selection.get_selected_rows()
        paths.reverse()
        for path in paths:
            node = model.get_iter(path)
            if node:
                model.remove(node)


# -------------------------------------------------------------------------
#
# ClipboardMultiTreeView class
#
# -------------------------------------------------------------------------
class ClipboardMultiTreeView(MultiTreeView):
    """
    TreeView that captures mouse events to make drag and drop work properly
    """

    def __init__(self, dbstate, uistate, title=None):
        super().__init__(uistate, "clipboard")
        self.dbstate = dbstate
        self.title = title if title else _("Clipboard")

    def on_button_press(self, widget, event):
        # Here we intercept mouse clicks on selected items so that we can
        # drag multiple items without the click selecting only one
        target = self.get_path_at_pos(int(event.x), int(event.y))
        if is_right_click(event):
            selection = widget.get_selection()
            store, paths = selection.get_selected_rows()
            if not paths:
                return
            tpath = paths[0] if paths else None
            node = store.get_iter(tpath) if tpath else None
            _ob = None
            if node:
                _ob = store.get_value(node, 1)
            self.newmenu = Gtk.Menu()
            popup = self.newmenu
            # ---------------------------
            if _ob:
                objclass, handle = _ob._objclass, _ob._handle
            else:
                objclass, handle = None, None
            if obj2class(objclass):
                if self.dbstate.db.method("has_%s_handle", objclass)(handle):
                    menu_item = Gtk.MenuItem(
                        label=_("See %s details", "the object")
                        % glocale.trans_objclass(objclass)
                    )
                    menu_item.connect(
                        "activate", lambda widget: self.edit_obj(objclass, handle)
                    )
                    popup.append(menu_item)
                    menu_item.show()
                    # ---------------------------
                    menu_item = Gtk.MenuItem(
                        label=_("Make %s active", "the object")
                        % glocale.trans_objclass(objclass)
                    )
                    menu_item.connect(
                        "activate",
                        lambda widget: self.uistate.set_active(handle, objclass),
                    )
                    popup.append(menu_item)
                    menu_item.show()
                # ---------------------------
                gids = set()
                for path in paths:
                    node = store.get_iter(path)
                    if node:
                        _ob = store.get_value(node, 1)
                        if _ob._objclass == objclass:
                            my_handle = _ob._handle
                            if self.dbstate.db.method("has_%s_handle", objclass)(
                                my_handle
                            ):
                                obj = self.dbstate.db.method(
                                    "get_%s_from_handle", objclass
                                )(my_handle)
                                gids.add(obj.gramps_id)
                if gids:
                    menu_item = Gtk.MenuItem(
                        label=_("Create Filter from %s selected...", "the object")
                        % glocale.trans_objclass(objclass)
                    )
                    menu_item.connect(
                        "activate",
                        lambda widget: make_filter(
                            self.dbstate, self.uistate, objclass, gids, title=self.title
                        ),
                    )
                    popup.append(menu_item)
                    menu_item.show()
            if popup.get_children():  # Show the popup menu:
                popup.popup_at_pointer(event)
            return True
        elif event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            model, paths = self.get_selection().get_selected_rows()
            for path in paths:
                node = model.get_iter(path)
                if node is not None:
                    _ob = model.get_value(node, 1)
                    objclass = _ob._objclass
                    handle = _ob._handle
                    self.edit_obj(objclass, handle)
            return True
        # otherwise:
        super().on_button_press(widget, event)

    def edit_obj(self, objclass, handle):
        from .editors import (
            EditPerson,
            EditEvent,
            EditFamily,
            EditSource,
            EditPlace,
            EditRepository,
            EditNote,
            EditMedia,
            EditCitation,
        )

        if obj2class(objclass):  # make sure it is an editable object
            if self.dbstate.db.method("has_%s_handle", objclass)(handle):
                g_object = self.dbstate.db.method("get_%s_from_handle", objclass)(
                    handle
                )
                try:
                    locals()["Edit" + objclass](
                        self.dbstate, self.uistate, [], g_object
                    )
                except WindowActiveError:
                    pass


def short(val, size=60):
    if len(val) > size:
        return "%s..." % val[0:size]
    return val


def place_title(db, event):
    return place_displayer.display_event(db, event)


def gen_del_obj(func, t):
    return lambda l: func(l, t)


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
def Clipboard(database, person, callback, parent=None):
    ClipboardWindow(database, parent)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale
from ....db import Database

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import HasGrampsId


# -------------------------------------------------------------------------
#
# HasIdOf
#
# -------------------------------------------------------------------------
class HasIdOf(HasGrampsId):
    """Rule that checks for a family with a specific Gramps ID"""

    name = _("Family with <Id>")
    description = _("Matches a family with a specified Gramps ID")

    def prepare(self, db: Database, user):
        data = db._get_raw_family_from_id_data(self.list[0])
        if data:
            self.selected_handles = set([data.handle])
        else:
            self.selected_handles = set([])



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....datehandler import parser
from .. import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Media
from ....db import Database


# -------------------------------------------------------------------------
#
# HasMedia
#
# -------------------------------------------------------------------------
class HasMedia(Rule):
    """Rule that checks for a media with a particular value"""

    labels = [
        _("Title:"),
        _("Type:"),
        _("Path:"),
        _("Date:"),
    ]
    name = _("Media objects matching parameters")
    description = _("Matches media objects with particular parameters")
    category = _("General filters")
    allow_regex = True

    def prepare(self, db: Database, user):
        self.date = None
        try:
            if self.list[3]:
                self.date = parser.parse(self.list[3])
        except:
            pass

    def apply_to_one(self, db: Database, obj: Media) -> bool:
        if not self.match_substring(0, obj.desc):
            return False

        if not self.match_substring(1, obj.mime):
            return False

        if not self.match_substring(2, obj.path):
            return False

        if self.date:
            if not obj.date.match(self.date):
                return False

        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Provide merge capabilities for citations.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib import (
    Person,
    Family,
    Event,
    Place,
    Media,
    Repository,
    Citation,
    Source,
    Note,
)
from ..db import DbTxn
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..errors import MergeError


# -------------------------------------------------------------------------
#
# MergeCitationQuery
#
# -------------------------------------------------------------------------
class MergeCitationQuery:
    """
    Create database query to merge two citations.
    """

    def __init__(self, dbstate, phoenix, titanic):
        self.database = dbstate.db
        self.phoenix = phoenix
        self.titanic = titanic

    def execute(self):
        """
        Merges to citations into a single citation.
        """
        new_handle = self.phoenix.get_handle()
        old_handle = self.titanic.get_handle()

        self.phoenix.merge(self.titanic)

        with DbTxn(_("Merge Citation"), self.database) as trans:
            self.database.commit_citation(self.phoenix, trans)
            for class_name, handle in self.database.find_backlink_handles(old_handle):
                if class_name == Person.__name__:
                    person = self.database.get_person_from_handle(handle)
                    assert person.has_citation_reference(old_handle)
                    person.replace_citation_references(old_handle, new_handle)
                    self.database.commit_person(person, trans)
                elif class_name == Family.__name__:
                    family = self.database.get_family_from_handle(handle)
                    assert family.has_citation_reference(old_handle)
                    family.replace_citation_references(old_handle, new_handle)
                    self.database.commit_family(family, trans)
                elif class_name == Event.__name__:
                    event = self.database.get_event_from_handle(handle)
                    assert event.has_citation_reference(old_handle)
                    event.replace_citation_references(old_handle, new_handle)
                    self.database.commit_event(event, trans)
                elif class_name == Place.__name__:
                    place = self.database.get_place_from_handle(handle)
                    assert place.has_citation_reference(old_handle)
                    place.replace_citation_references(old_handle, new_handle)
                    self.database.commit_place(place, trans)
                elif class_name == Media.__name__:
                    obj = self.database.get_media_from_handle(handle)
                    assert obj.has_citation_reference(old_handle)
                    obj.replace_citation_references(old_handle, new_handle)
                    self.database.commit_media(obj, trans)
                elif class_name == Repository.__name__:
                    repository = self.database.get_repository_from_handle(handle)
                    assert repository.has_citation_reference(old_handle)
                    repository.replace_citation_references(old_handle, new_handle)
                    self.database.commit_repository(repository, trans)
                elif class_name == Citation.__name__:
                    citation = self.database.get_citation_from_handle(handle)
                    assert citation.has_citation_reference(old_handle)
                    citation.replace_citation_references(old_handle, new_handle)
                    self.database.commit_citation(citation, trans)
                elif class_name == Source.__name__:
                    source = self.database.get_source_from_handle(handle)
                    assert source.has_citation_reference(old_handle)
                    source.replace_citation_references(old_handle, new_handle)
                    self.database.commit_source(source, trans)
                elif class_name == Note.__name__:
                    note = self.database.get_note_from_handle(handle)
                    assert note.has_handle_reference("Citation", old_handle)
                    note.replace_handle_reference("Citation", old_handle, new_handle)
                    self.database.commit_note(note, trans)
                else:
                    raise MergeError(
                        "Encounter an object of type %s that has "
                        "a citation reference." % class_name
                    )
            self.database.remove_citation(old_handle, trans)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import RegExpIdBase

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Family
from ....db import Database


# -------------------------------------------------------------------------
#
# HasNameOf
#
# -------------------------------------------------------------------------
class MotherHasIdOf(RegExpIdBase):
    """Rule that checks for a person with a specific Gramps ID"""

    labels = [_("Person ID:")]
    name = _("Families having mother with Id containing <text>")
    description = _("Matches families whose mother has a specified " "Gramps ID")
    category = _("Mother filters")

    def apply_to_one(self, db: Database, family: Family) -> bool:  # type: ignore[override]
        mother_handle = family.mother_handle
        if mother_handle:
            mother = db.get_person_from_handle(mother_handle)
            if mother:
                return super().apply_to_one(db, mother)
            else:
                return False
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib import Person
from ...db import Database


# -------------------------------------------------------------------------
# "People having notes that contain a substring"
# -------------------------------------------------------------------------
class HasNoteSubstrBase(Rule):
    """People having notes containing <substring>."""

    labels = [_("Substring:")]
    name = "Objects having notes containing <substring>"
    description = "Matches objects whose notes contain text matching a " "substring"
    category = _("General filters")

    def apply_to_one(self, db, person: Person) -> bool:
        notelist = person.note_list
        for notehandle in notelist:
            note = db.get_note_from_handle(notehandle)
            n = str(note.text)
            if n.upper().find(self.list[0].upper()) != -1:
                return True
        return False



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
#               2009-2011  Gary Burton
#               2025       Steve Youngs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
from __future__ import annotations

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from ..managedwindow import ManagedWindow
from ..filters import SearchBar
from ..glade import Glade
from ..widgets.interactivesearchbox import InteractiveSearchBox
from ..display import display_help
from gramps.gen.const import URL_MANUAL_PAGE
from gramps.gen.filters import GenericFilter
from gramps.gui.widgets.persistenttreeview import PersistentTreeView
from gramps.gui.widgets.multitreeview import MultiTreeView


# -------------------------------------------------------------------------
#
# SelectEvent
#
# -------------------------------------------------------------------------
class BaseSelector(ManagedWindow):
    """Base class for the selectors, showing a dialog from which to select
    one of the primary objects
    """

    NONE = -1
    TEXT = 0
    MARKUP = 1
    IMAGE = 2

    WIKI_HELP_PAGE: str
    WIKI_HELP_SEC: str

    def __init__(
        self,
        dbstate,
        uistate,
        track=[],
        search_or_filter: (
            tuple[int, str] | GenericFilter | None
        ) = None,  # tuple[search_index, search_string]
        skip=set(),
        show_search_bar: bool = True,
        default=None,
        allow_multiple_selection: bool = False,
    ):
        """Set up the dialog with the dbstate and uistate, track of parent
        windows for ManagedWindow, initial filter for the model, skip with
        set of handles to skip in the view, and search_bar to show the
        SearchBar at the top or not.
        """
        # Set window title, some selectors may set self.title in their __init__
        if not hasattr(self, "title"):
            self.title = self.get_window_title()

        ManagedWindow.__init__(self, uistate, track, self)

        self.renderer = Gtk.CellRendererText()
        self.track_ref_for_deletion("renderer")
        self.renderer.set_property("ellipsize", Pango.EllipsizeMode.END)

        self.db = dbstate.db
        self.tree = None
        self.model = None

        self.glade = Glade()

        window = self.glade.toplevel
        title_label = self.glade.get_object("title")
        vbox = self.glade.get_object("select_person_vbox")
        objectlist = self.glade.get_object("plist")
        _config_name = self.get_config_name()
        self.allow_multiple_selection = allow_multiple_selection
        self.tree = (
            MultiTreeView(self.uistate, _config_name)
            if self.allow_multiple_selection
            else PersistentTreeView(self.uistate, _config_name)
        )
        if self.allow_multiple_selection:
            self.tree.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        scrolledwindow = self.glade.get_object("scrolledwindow33")
        scrolledwindow.remove(objectlist)
        scrolledwindow.add(self.tree)
        self.tree.set_headers_visible(True)
        self.tree.set_headers_clickable(True)
        self.tree.connect("row-activated", self._on_row_activated)
        self.tree.grab_focus()
        self.define_help_button(
            self.glade.get_object("help"), self.WIKI_HELP_PAGE, self.WIKI_HELP_SEC
        )

        # connect to signal for custom interactive-search
        self.searchbox = InteractiveSearchBox(self.tree)
        self.tree.connect("key-press-event", self.searchbox.treeview_keypress)

        # add the search bar
        self.search_bar = SearchBar(dbstate, uistate, self.build_tree)
        search_box = self.search_bar.build()
        self.setup_searches()
        if isinstance(search_or_filter, GenericFilter):
            self.search_bar.append_filter(search_or_filter.name, search_or_filter)
            self.search_bar.search_list.set_active(
                self.search_bar.search_model.iter_n_children(None) - 1
            )
        elif isinstance(search_or_filter, tuple):
            self.search_bar.search_list.set_active(search_or_filter[0])
            self.search_bar.search_text.set_text(search_or_filter[1])
        vbox.pack_start(search_box, False, False, 0)
        vbox.reorder_child(search_box, 1)

        self.set_window(window, title_label, self.title)

        # set up sorting
        self.sort_col = 0
        self.setupcols = True
        self.columns: list[Gtk.TreeViewColumn] = []
        self.sortorder = Gtk.SortType.ASCENDING

        self.skip_list = skip
        self.selection = self.tree.get_selection()
        self.track_ref_for_deletion("selection")

        self._local_init()
        self._set_size()

        self.show()
        # show or hide search bar?
        self.set_show_search_bar(show_search_bar)
        while Gtk.events_pending():
            Gtk.main_iteration()
        self.build_tree()
        loading = self.glade.get_object("loading")
        loading.hide()

        if default:
            self.goto_handle(default)

    def goto_handle(self, handle):
        """
        Goto the correct row.
        """
        iter_ = self.model.get_iter_from_handle(handle)
        if iter_:
            if not (self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY):
                # Expand tree
                parent_iter = self.model.iter_parent(iter_)
                if parent_iter:
                    parent_path = self.model.get_path(parent_iter)
                    if parent_path:
                        parent_path_list = parent_path.get_indices()
                        for i, value in enumerate(parent_path_list):
                            expand_path = Gtk.TreePath(
                                tuple([x for x in parent_path_list[: i + 1]])
                            )
                            self.tree.expand_row(expand_path, False)

            # Select active object
            path = self.model.get_path(iter_)
            self.selection.unselect_all()
            self.selection.select_path(path)
            self.tree.scroll_to_cell(path, None, 1, 0.5, 0)
        else:
            self.selection.unselect_all()

    def add_columns(self, tree):
        tree.set_fixed_height_mode(True)
        titles = self.get_column_titles()
        for ix, item in enumerate(titles):
            if item[2] == BaseSelector.NONE:
                continue
            elif item[2] == BaseSelector.TEXT:
                column = Gtk.TreeViewColumn(item[0], self.renderer, text=item[3])
            elif item[2] == BaseSelector.MARKUP:
                column = Gtk.TreeViewColumn(item[0], self.renderer, markup=item[3])
            column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
            column.set_fixed_width(item[1])
            column.set_resizable(True)
            # connect click
            column.connect("clicked", self.column_clicked, ix)
            column.set_clickable(True)
            ##column.set_sort_column_id(ix) # model has its own sort implemented
            self.columns.append(column)
            tree.append_column(column)

    def build_menu_names(self, obj):
        return (self.title, None)

    def get_selected_ids(self):
        mlist = []
        self.selection.selected_foreach(self.select_function, mlist)
        return mlist

    def first_selected(self):
        """first selected entry in the Selector tree"""
        mlist = []
        self.selection.selected_foreach(self.select_function, mlist)
        return mlist[0] if mlist else None

    def select_function(self, store, path, iter_, id_list):
        handle = store.get_handle_from_iter(iter_)
        id_list.append(handle)

    def run(self):
        val = self.window.run()
        result = None
        if val == Gtk.ResponseType.OK:
            handle_list = self.get_selected_ids()
            if handle_list:
                if self.allow_multiple_selection:
                    # always return a list, but may be length 0
                    result = [
                        self.get_from_handle_func()(handle)
                        for handle in handle_list
                        if handle
                    ]
                else:
                    # return None or a valid handle
                    if handle_list[0]:
                        result = self.get_from_handle_func()(handle_list[0])
            self.close()
        elif val != Gtk.ResponseType.DELETE_EVENT:
            self.close()
        return result

    def _on_row_activated(self, treeview, path, view_col):
        self.window.response(Gtk.ResponseType.OK)

    def _local_init(self):
        # define selector-specific init routine
        pass

    def get_window_title(self):
        assert False, "Must be defined in the subclass"

    def get_model_class(self):
        assert False, "Must be defined in the subclass"

    def get_column_titles(self):
        """
        Defines the columns to show in the selector. Must be defined in the
        subclasses.
        :returns: a list of tuples with four entries. The four entries should
                be:
                0: column header string,
                1: column width,
                2: TEXT, MARKUP or IMAGE,
                3: column in the model that must be used.
        """
        raise NotImplementedError

    def get_from_handle_func(self):
        assert False, "Must be defined in the subclass"

    def set_show_search_bar(self, value):
        """make the search bar at the top shown"""
        self.show_search_bar = value
        if not self.search_bar:
            return
        if self.show_search_bar:
            self.search_bar.show()
        else:
            self.search_bar.hide()

    def column_order(self):
        """
        returns a tuple indicating the column order of the model
        """
        return [(1, row[3], row[1], row[0]) for row in self.get_column_titles()]

    def exact_search(self):
        """
        Returns a tuple indicating columns requiring an exact search
        """
        return ()

    def setup_searches(self):
        """
        Builds the default searches and add them to the search bar.
        """
        cols = [
            (pair[3], pair[1], pair[1] in self.exact_search())
            for pair in self.column_order()
            if pair[0]
        ]
        self.search_bar.setup_searches(cols)

    def build_tree(self):
        """
        Builds the selection people see in the Selector
        """
        filter_info = self.search_bar.get_value()
        if not filter_info[0]:
            filter_info = (
                filter_info[0],
                filter_info[1],
                filter_info[1][0] in self.exact_search(),
            )

        if self.model:
            sel = self.first_selected()
        else:
            sel = None

        # set up cols the first time
        if self.setupcols:
            self.add_columns(self.tree)

        # reset the model with correct sorting
        self.clear_model()
        self.model = self.get_model_class()(
            self.db,
            self.uistate,
            self.sort_col,
            self.sortorder,
            sort_map=self.column_order(),
            skip=self.skip_list,
            search=filter_info,
        )

        self.tree.set_model(self.model)
        self.tree.restore_column_size()

        # sorting arrow in column header (not on start, only on click)
        if not self.setupcols:
            for i, column in enumerate(self.columns):
                enable_sort_flag = i == self.sort_col
                column.set_sort_indicator(enable_sort_flag)
            self.columns[self.sort_col].set_sort_order(self.sortorder)

        # set the search column to be the sorted column
        search_col = self.column_order()[self.sort_col][1]
        self.tree.set_search_column(search_col)

        self.setupcols = False
        if sel:
            self.goto_handle(sel)

    def get_config_name(self):
        assert False, "Must be defined in the subclass"

    def column_clicked(self, obj, data):
        if self.sort_col != data:
            self.sortorder = Gtk.SortType.ASCENDING
            self.sort_col = data
        else:
            if (
                self.columns[data].get_sort_order() == Gtk.SortType.DESCENDING
                or not self.columns[data].get_sort_indicator()
            ):
                self.sortorder = Gtk.SortType.ASCENDING
            else:
                self.sortorder = Gtk.SortType.DESCENDING
        self.build_tree()

        return True

    def clear_model(self):
        if self.model:
            self.tree.set_model(None)
            if hasattr(self.model, "destroy"):
                self.model.destroy()
            self.model = None

    def _cleanup_on_exit(self):
        """Unset all things that can block garbage collection.
        Finalize rest
        """
        self.clear_model()
        self.db = None
        self.tree = None
        self.columns = None
        self.search_bar.destroy()

    def close(self, *obj):
        ManagedWindow.close(self)
        self._cleanup_on_exit()

    def define_help_button(self, button, webpage="", section=""):
        """Setup to deal with help button"""
        button.connect("clicked", lambda x: display_help(webpage, section))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2015       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....display.place import displayer

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Place
from ....db import Database


# -------------------------------------------------------------------------
#
# HasTitle
#
# -------------------------------------------------------------------------
class HasTitle(Rule):
    """
    Rule that checks for a place with a title
    """

    labels = [_("Title:")]
    name = _("Places matching a title")
    description = _("Matches places with a particular title")
    category = _("General filters")
    allow_regex = True

    def apply_to_one(self, db: Database, place: Place) -> bool:
        if not self.match_substring(0, displayer.display(db, place)):
            return False
        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Zsolt Foldvari
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"ShortlistComboEntry class."

__all__ = ["ShortlistComboEntry"]

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.shortlistcomboentry")

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .validatedcomboentry import ValidatedComboEntry

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

_GTYPE = {
    str: GObject.TYPE_STRING,
    int: GObject.TYPE_INT64,
    float: GObject.TYPE_FLOAT,
}

(
    COLUMN_ITEM,
    COLUMN_IS_SEP,
) = list(range(2))


# -------------------------------------------------------------------------
#
# ShortlistComboEntry class
#
# -------------------------------------------------------------------------
class ShortlistComboEntry(ValidatedComboEntry):
    """A ComboboxEntry class with optional shortlist."""

    __gtype_name__ = "ShortlistComboEntry"

    def __init__(self):
        pass

    def init(self, items, shortlist=True, validator=None):
        if not items:
            raise ValueError

        data_type = items[0].__class__
        gtype = _GTYPE.get(data_type, GObject.TYPE_PYOBJECT)

        # create the model and insert the items
        model = Gtk.ListStore(gtype, GObject.TYPE_BOOLEAN)
        maxlen = -1
        for item in items:
            if len(str(item)) > maxlen:
                maxlen = len(str(item))
            model.append((item, False))

        width = -1  # default width
        if 1 < maxlen < 4:
            width = 4
        elif 1 < maxlen < 10:
            width = maxlen + 1
        ValidatedComboEntry.__init__(
            self, data_type, model, COLUMN_ITEM, validator, width=width
        )
        if shortlist:
            self._shortlist = []
            self.connect("changed", self._on_combobox_changed)

        self.set_row_separator_func(self._is_row_separator, None)

    def _on_combobox_changed(self, combobox):
        if self._internal_change:
            return

        if self.get_active_iter():
            model = self.get_model()

            # if first item on shortlist insert a separator row
            if not self._shortlist:
                model.prepend((None, True))

            # remove the existing shortlist from the model
            iter = model.get_iter_first()
            for n in self._shortlist:
                model.remove(iter)

            # update shortlist
            if self._active_data in self._shortlist:
                self._shortlist.remove(self._active_data)
            self._shortlist.append(self._active_data)
            self._shortlist = self._shortlist[-5:]

            # prepend shortlist to model
            for data in self._shortlist:
                model.prepend((data, False))

    def _is_row_separator(self, model, iter, data=None):
        return model.get_value(iter, COLUMN_IS_SEP)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2017-      Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Package providing filter rules for Gramps.
"""

from ._allplaces import AllPlaces
from ._hascitation import HasCitation
from ._hasgallery import HasGallery
from ._hasidof import HasIdOf
from ._regexpidof import RegExpIdOf
from ._hasnote import HasNote
from ._hasnoteregexp import HasNoteRegexp
from ._hasnotematchingsubstringof import HasNoteMatchingSubstringOf
from ._hasnotetype import HasNoteType
from ._hasreferencecountof import HasReferenceCountOf
from ._hassourcecount import HasSourceCount
from ._hassourceof import HasSourceOf
from ._placeprivate import PlacePrivate
from ._matchesfilter import MatchesFilter
from ._hasplace import HasPlace
from ._hasdata import HasData
from ._hasnolatorlon import HasNoLatOrLon
from ._inlatlonneighborhood import InLatLonNeighborhood
from ._isenclosedby import IsEnclosedBy
from ._matcheseventfilter import MatchesEventFilter
from ._matchessourceconfidence import MatchesSourceConfidence
from ._changedsince import ChangedSince
from ._hastag import HasTag
from ._hastitle import HasTitle
from ._withinarea import WithinArea

editor_rule_list = [
    AllPlaces,
    HasCitation,
    HasGallery,
    HasIdOf,
    RegExpIdOf,
    HasNote,
    HasNoteRegexp,
    HasNoteType,
    HasReferenceCountOf,
    HasSourceCount,
    HasSourceOf,
    PlacePrivate,
    MatchesFilter,
    MatchesSourceConfidence,
    HasData,
    HasNoLatOrLon,
    InLatLonNeighborhood,
    MatchesEventFilter,
    ChangedSince,
    HasTag,
    HasTitle,
    WithinArea,
    IsEnclosedBy,
]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011  Jerome Rapinat
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
# Filters/Rules/Person/_MatchesSourceConfidence.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .._matchessourceconfidencebase import MatchesSourceConfidenceBase


# -------------------------------------------------------------------------
# "Confidence level"
# -------------------------------------------------------------------------
class MatchesSourceConfidence(MatchesSourceConfidenceBase):
    """Media matching a specific confidence level on its 'direct' source references"""

    labels = [_("Confidence level:")]
    name = _("Place with a direct source >= <confidence level>")
    description = _(
        "Matches places with at least one direct source with confidence level(s)"
    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Provide merge capabilities for repositories.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib import Source, Note
from ..db import DbTxn
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..errors import MergeError


# -------------------------------------------------------------------------
#
# MergeRepoQuery
#
# -------------------------------------------------------------------------
class MergeRepositoryQuery:
    """
    Create database query to merge two repositories.
    """

    def __init__(self, dbstate, phoenix, titanic):
        self.database = dbstate.db
        self.phoenix = phoenix
        self.titanic = titanic

    def execute(self):
        """
        Merges two repositories into a single repository.
        """
        new_handle = self.phoenix.get_handle()
        old_handle = self.titanic.get_handle()

        self.phoenix.merge(self.titanic)

        with DbTxn(_("Merge Repositories"), self.database) as trans:
            self.database.commit_repository(self.phoenix, trans)
            for class_name, handle in self.database.find_backlink_handles(old_handle):
                if class_name == Source.__name__:
                    source = self.database.get_source_from_handle(handle)
                    assert source.has_handle_reference("Repository", old_handle)
                    source.replace_repo_references(old_handle, new_handle)
                    self.database.commit_source(source, trans)
                elif class_name == Note.__name__:
                    note = self.database.get_note_from_handle(handle)
                    assert note.has_handle_reference("Repository", old_handle)
                    note.replace_handle_reference("Repository", old_handle, new_handle)
                    self.database.commit_note(note, trans)
                else:
                    raise MergeError(
                        "Encounter an object of type %s that has "
                        "a repository reference." % class_name
                    )
            self.database.remove_repository(old_handle, trans)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Any
from ...db import Database


# -------------------------------------------------------------------------
#
# HasSource
#
# -------------------------------------------------------------------------
class HasSourceBase(Rule):
    """Rule that checks for a source with a particular value"""

    labels = ["Title:", "Author:", "Abbreviation:", "Publication:"]
    name = "Sources matching parameters"
    description = "Matches sources with particular parameters"
    category = _("Citation/source filters")
    allow_regex = True

    def apply_to_one(self, db: Database, source: Any) -> bool:
        if not self.match_substring(0, source.title):
            return False

        if not self.match_substring(1, source.author):
            return False

        if not self.match_substring(2, source.abbrev):
            return False

        if not self.match_substring(3, source.pubinfo):
            return False

        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Brian G. Matherly
# Copyright (C) 2008  Jerome Rapinat
# Copyright (C) 2008  Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#
# gen.filters.rules/_HasGalleryBase.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib.mediabase import MediaBase
from ...db import Database


# -------------------------------------------------------------------------
# "Objects that have images"
# -------------------------------------------------------------------------
class HasGalleryBase(Rule):
    """Objects who have Media Object"""

    labels = [_("Number of instances:"), _("Number must be:")]
    name = "Object with <count> Media references"
    description = "Matches objects with certain number of items in the gallery"
    category = _("General filters")

    def prepare(self, db: Database, user):
        # things we want to do just once, not for every handle
        if self.list[1] == "less than":
            self.count_type = 0
        elif self.list[1] == "greater than":
            self.count_type = 2
        else:
            self.count_type = 1  # "equal to"

        self.userSelectedCount = int(self.list[0])

    def apply_to_one(self, db: Database, obj: MediaBase) -> bool:
        count = len(obj.media_list)
        if self.count_type == 0:  # "less than"
            return count < self.userSelectedCount
        elif self.count_type == 2:  # "greater than"
            return count > self.userSelectedCount
        # "equal to"
        return count == self.userSelectedCount



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2015,2017  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <https://www.gnu.org/licenses/>.
#

"""
Place name class for Gramps
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from .const import DIFFERENT, EQUAL, IDENTICAL
from .datebase import DateBase
from .secondaryobj import SecondaryObject

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PlaceName
#
# -------------------------------------------------------------------------
class PlaceName(SecondaryObject, DateBase):
    """
    Place name class.

    This class is for keeping information about place names.
    """

    def __init__(self, source=None, **kwargs):
        """
        Create a new PlaceName instance, copying from the source if present.
        """
        DateBase.__init__(self, source)
        if source:
            self.value = source.value
            self.lang = source.lang
        else:
            self.value = ""
            self.lang = ""
        for key, value in kwargs.items():
            if key in ["value", "lang"]:
                setattr(self, key, value)
            else:
                raise AttributeError(f"PlaceName does not have property '{key}'")

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (self.value, DateBase.serialize(self), self.lang)

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.value, date, self.lang = data
        DateBase.unserialize(self, date)
        return self

    @classmethod
    def get_schema(cls):
        """
        Returns the JSON Schema for this class.

        :returns: Returns a dict containing the schema.
        :rtype: dict
        """
        # pylint: disable=import-outside-toplevel
        from .date import Date

        return {
            "type": "object",
            "title": _("Place Name"),
            "properties": {
                "_class": {"enum": [cls.__name__]},
                "value": {"type": "string", "title": _("Text")},
                "date": Date.get_schema(),
                "lang": {"type": "string", "title": _("Language")},
            },
        }

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return [self.value]

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return []

    def get_citation_child_list(self):
        """
        Return the list of child secondary objects that may refer citations.

        :returns: Returns the list of child secondary child objects that may
                  refer citations.
        :rtype: list
        """
        return []

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        :returns: Returns the list of child secondary child objects that may
                  refer notes.
        :rtype: list
        """
        return []

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: Returns the list of (classname, handle) tuples for referenced
                  objects.
        :rtype: list
        """
        return []

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through their
        children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return []

    def is_equivalent(self, other):
        """
        Return if this eventref is equivalent, that is agrees in handle and
        role, to other.

        :param other: The eventref to compare this one to.
        :type other: PlaceName
        :returns: Constant indicating degree of equivalence.
        :rtype: int
        """
        if (
            self.value != other.value
            or self.date != other.date
            or self.lang != other.lang
        ):
            return DIFFERENT
        if self.is_equal(other):
            return IDENTICAL
        return EQUAL

    def __eq__(self, other):
        return self.is_equal(other)

    def __ne__(self, other):
        return not self.is_equal(other)

    def set_value(self, value):
        """
        Set the name for the PlaceName instance.
        """
        self.value = value

    def get_value(self):
        """
        Return the name for the PlaceName instance.
        """
        return self.value

    def set_language(self, lang):
        """
        Set the language for the PlaceName instance.
        """
        self.lang = lang

    def get_language(self):
        """Return the language for the PlaceName instance."""
        return self.lang
